import os
from collections import defaultdict
from typing import DefaultDict, Dict, Generator

from fastapi import FastAPI

from kess import env, imports
from kess.function import Function


class App(FastAPI):
    production: bool
    functions_folder: str
    route_prefix: str
    functions: DefaultDict[str, Dict[str, Function]]

    def __init__(self, *args, **kwargs):
        self.production = kwargs.pop("production", env.PRODUCTION)
        self.functions_folder = kwargs.pop("functions_folder", env.FUNCTIONS_FOLDER)
        self.route_prefix = kwargs.pop("router_prefix", env.ROUTER_PREFIX)
        self.functions = defaultdict(dict)
        super().__init__(*args, **kwargs)

    def scan_prod_functions(
        self, functions_folder: str
    ) -> Generator[(str, str, Function)]:
        for function_name in os.listdir(functions_folder):
            function_path = os.path.join(functions_folder, function_name)
            if os.path.isdir(function_path):
                for function_file_name in os.listdir(function_path):
                    function_file_path = os.path.join(function_path, function_file_name)
                    if os.path.isfile(function_file_path):
                        function_version = os.path.splitext(function_file_name)[0]
                        module = imports.import_module(os.path.splitext(function_file_path)[0].replace(os.path.sep, "."))
                        yield function_name, function_version, module.fn

    def scan_dev_functions(
        self, functions_folder: str
    ) -> Generator[(str, str, Function)]:
        for function_file_name in os.listdir(functions_folder):
            function_file_path = os.path.join(functions_folder, function_file_name)
            if os.path.isfile(function_file_path):
                function_name = os.path.splitext(function_file_name)[0]
                function_version = "latest"
                module = imports.import_module(os.path.splitext(function_file_path)[0].replace(os.path.sep, "."))
                yield function_name, function_version, module.fn

    def scan_functions(
        self, functions_folder: str, production: bool = False
    ) -> Generator[(str, str, Function)]:
        scanner = self.scan_prod_functions if production else self.scan_dev_functions
        yield from scanner(functions_folder)

    def setup_function(self, fn: Function, name: str, version: str, prefix: str = ""):
        fn.state.app = self
        self.mount(f"{prefix}/{name}/{version}", fn)
        self.functions[name][version] = fn

    def setup_functions(
        self, functions_folder: str, production: bool = False, prefix: str = ""
    ):
        for name, version, fn in self.scan_functions(
            functions_folder, production=production
        ):
            self.setup_function(fn, name, version, prefix=prefix)

    def setup(self):
        super().setup()
        self.setup_functions(
            self.functions_folder, production=self.production, prefix=self.route_prefix
        )



import asyncio
from concurrent.futures import process

_executor = None


def executor(*args, **kwargs) -> process.ProcessPoolExecutor:
    global _executor
    if not _executor:
        _executor = process.ProcessPoolExecutor(*args, **kwargs)
    return _executor


async def run_in_executor(executor, *args, **kwargs):
    return await asyncio.get_event_loop().run_in_executor(executor, *args, **kwargs)


async def run_in_process(*args, **kwargs):
    return await run_in_executor(executor(), *args, **kwargs)



from typing import Callable

from fastapi import FastAPI


class Function(FastAPI):
    def handle(self, fn) -> Callable:
        return self.post("/")(fn)

    def h(self, fn) -> Callable:
        return self.handle(fn)



import importlib
import importlib.util
import uuid
import sys
from types import ModuleType

import_module = importlib.import_module
reload = importlib.reload


def import_file(file_path: str, module_name: str = None) -> ModuleType:
    module_name = module_name or uuid.uuid4().hex
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module.__name__] = module
    spec.loader.exec_module(module)
    return module



import os

PRODUCTION = bool(os.getenv("KESS_PRODUCTION"))
HOST = os.getenv("KESS_HOST", "0.0.0.0")
PORT = int(os.getenv("KESS_PORT", "8000"))
FUNCTIONS_FOLDER = os.getenv("KESS_FUNCTIONS_FOLDER", "functions")
ROUTER_PREFIX = os.getenv("KESS_ROUTE_PREFIX", "")
